--****
-- == Pipe Input/Output
--
-- <<LEVELTOC level=2 depth=4>>
--
-- === Notes
-- Due to a bug, Euphoria does not handle STDERR properly
-- STDERR cannot captured for Euphoria programs (other programs will work fully)
-- The IO functions currently work with file handles, a future version might wrap them in streams
-- so that they can be used directly alongside other file/socket/other-streams with a
-- stream_select() function.
--

include dll.e
include machine.e
include lib/_dll_.e
include lib/_machine_.e

constant
  kernel32 = open_dll("kernel32.dll"),
  --iGetExitCodeProcess=define_c_func(kernel32,"GetExitCodeProcess",{C_UINT,C_POINTER},C_INT),
  iCreatePipe = define_c_func(kernel32,"CreatePipe",{C_POINTER, C_POINTER, C_POINTER, C_LONG},C_USHORT),
  iReadFile = define_c_func(kernel32,"ReadFile",{C_POINTER, C_POINTER, C_LONG, C_POINTER, C_POINTER}, C_USHORT),
  iWriteFile = define_c_func(kernel32,"WriteFile",{C_POINTER, C_POINTER, C_LONG, C_POINTER, C_POINTER}, C_USHORT),
  iCloseHandle = define_c_func(kernel32,"CloseHandle",{C_POINTER}, C_USHORT),
  iTerminateProcess=define_c_func(kernel32,"TerminateProcess",{C_POINTER, C_UINT}, C_USHORT),
  iGetLastError = define_c_func(kernel32,"GetLastError",{},C_LONG),
  iGetStdHandle = define_c_func(kernel32,"GetStdHandle",{C_LONG}, C_LONG),
  iSetHandleInformation = define_c_func(kernel32,"SetHandleInformation",{C_POINTER, C_LONG, C_LONG}, C_USHORT),
  iCreateProcess = define_c_func(kernel32,"CreateProcessA",{C_POINTER, C_POINTER, C_POINTER,
    C_POINTER, C_USHORT, C_LONG, C_POINTER, C_POINTER, C_POINTER, C_POINTER},C_USHORT)

constant
  PIPE_WRITE_HANDLE = 1, PIPE_READ_HANDLE=2,
  HANDLE_FLAG_INHERIT=1,
  STARTF_USESHOWWINDOW = 1,
  STARTF_USESTDHANDLES = 256,
  FAIL = 0

constant
  SA_SIZE = 12,
  SUIdwFlags = 44, 
  SUIhStdInput = 56, 
  STARTUPINFO_SIZE = 68,
  PROCESS_INFORMATION_SIZE = 16

--****
-- === Accessor Constants

global constant
  --** Child processes standard input
  STDIN=1,
  --** Child processes standard output
  STDOUT=2,
  --** Child processes standard error
  STDERR=3,
  --** Process ID
  PID=4

--
global constant
  --** Set of pipes that are for the use of the parent
  PARENT=1,
  --** Set of pipes that are given to the child - should not be used by the parent
  CHILD=2

--
atom os_errno
os_errno = 0

-- Common functions
function get_errno()
  return c_func(iGetLastError,{})
end function

--****
-- === Opening/Closing

--**
-- Process Type

global type process(object o)
  if atom(o) then
    return 0
  end if

  if length(o) != 4 then
    return 0
  end if

  return 1
end type

--**
-- Close handle fd
--
-- Returns:
--   An **integer**, 0 on success, -1 on failure
--
-- Example 1:
-- <eucode>
-- integer status = pipeio:close(p[STDIN])
-- </eucode>
--

global function close(atom fd)
  atom ret

  ret=c_func(iCloseHandle,{fd})
  if ret = FAIL then
    os_errno = get_errno()
    return -1
  end if
  
  return 0
end function

--**
-- Close pipes and kill process p with signal signal (default 15)
--
-- Comments:
--   Signal is ignored on Windows.
--
-- Example 1:
-- <eucode>
-- kill(p)
-- </eucode>
--

global procedure kill(process p, atom signal)
  void = close(p[STDIN])
  void = close(p[STDOUT])
  void = close(p[STDERR])
  
  void = c_func(iTerminateProcess,{p[PID],signal and 0})
end procedure

function os_pipe()
  sequence handles
  atom ret, psaAttrib, phWriteToPipe, phReadFromPipe
  
  psaAttrib = allocate(SA_SIZE + 2 * size_of( C_POINTER ) )
  poke4( psaAttrib, SA_SIZE )
  poke_pointer(psaAttrib + size_of( C_POINTER),{0,1})
  
  phWriteToPipe = psaAttrib + SA_SIZE
  phReadFromPipe = psaAttrib + SA_SIZE + size_of( C_POINTER )
  
  ret = c_func(iCreatePipe,{phReadFromPipe,phWriteToPipe,psaAttrib,0})
  handles = peek_pointer({phWriteToPipe,2})
  
  free(psaAttrib)
  
  if ret = FAIL then
    os_errno = get_errno()
    return -1
  end if
  
  return handles
end function

--****
-- === Read/Write Process

--**
-- Read ##bytes## bytes from handle ##fd##
--
-- Returns:
--   A **sequence**, containing data, an empty sequence on EOF or an error code.
--   Similar to [[:get_bytes]].
--
-- Example 1:
-- <eucode>
-- sequence data=read(p[STDOUT],256)
-- </eucode>
--

global function read(atom fd, integer bytes)
  sequence data
  atom ret, ReadCount, buf, pReadCount

  if bytes=0 then return "" end if

  buf = allocate(bytes)
  
  pReadCount=allocate(4)
  ret = c_func(iReadFile,{fd,buf,bytes,pReadCount,0})
  ReadCount=peek4u(pReadCount)
  free(pReadCount)
  
  if ret = FAIL then
    os_errno = get_errno()
    free(buf)
    return ""
  end if
  
  data=peek({buf,ReadCount})
  free(buf)
  
  return data
end function

--**
-- Write ##bytes## to handle ##fd##
--
-- Returns:
--   An **integer**, number of bytes written, or -1 on error
--
-- Example 1:
-- <eucode>
-- integer bytes_written = write(p[STDIN],"Hello World!")
-- </eucode>
--

global function write(atom fd, sequence str)
  atom buf, ret, WrittenCount, pWrittenCount
  
  buf = allocate_string(str)
  pWrittenCount=allocate(4)
  ret=c_func(iWriteFile,{fd,buf,length(str),pWrittenCount,0})
  WrittenCount=peek4u(pWrittenCount)
  free(pWrittenCount)
  
  free(buf)
  
  if ret = FAIL then
    os_errno = get_errno()
    return -1
  end if

  return WrittenCount
end function

procedure error()
  crash_message(sprintf("Errno = %d", os_errno))
end procedure

--**
-- Get error no from last call to a pipe function
--
-- Comments:
--   Value returned will be OS-specific, and is not always set on Windows at least
--
-- Example 1:
-- <eucode>
-- integer error = error_no()
-- </eucode>
--
global function error_no()
  return os_errno
end function

function GetStdHandle(atom device)
  return c_func(iGetStdHandle,{device})
end function

function SetHandleInformation(atom hObject, atom dwMask, atom dwFlags)
  return c_func(iSetHandleInformation,{hObject,dwMask,dwFlags})
end function

procedure CloseAllHandles(sequence handles)
   for i = 1 to length(handles) do
     void = close(handles[i])
   end for
end procedure

function CreateProcess(sequence CommandLine,sequence StdHandles)
  object fnVal
  atom pPI, pSUI, pCmdLine
  sequence ProcInfo

  pCmdLine = allocate_string(CommandLine)
  
  pPI = allocate(PROCESS_INFORMATION_SIZE)
  mem_set(pPI,0,PROCESS_INFORMATION_SIZE)
  
  pSUI = allocate(STARTUPINFO_SIZE)
  mem_set( pSUI, 0, STARTUPINFO_SIZE)
  poke4( pSUI, STARTUPINFO_SIZE)
  poke4( pSUI + SUIdwFlags, or_bits( STARTF_USESTDHANDLES, STARTF_USESHOWWINDOW ) )
  poke_pointer( pSUI + SUIhStdInput, StdHandles)
  
  fnVal = c_func(iCreateProcess,{0,pCmdLine,0,0,1,0,0,0,pSUI,pPI})
  free(pCmdLine)
  free(pSUI)
  ProcInfo = peek4u({pPI,4})
  free(pPI)
  if not fnVal then
    return 0
  end if
  return ProcInfo
end function

--**
-- Create pipes for inter-process communication
--
-- Returns:
--   A **handle**, process handles { {parent side pipes},{child side pipes} }
--
-- Example 1:
-- <eucode>
-- object p = exec("dir", create())
-- </eucode>
--
global function create()
  atom hChildStdInRd,hChildStdOutWr, hChildStdErrWr,
       hChildStdInWr, hChildStdOutRd,hChildStdErrRd

  object StdInPipe, StdOutPipe, StdErrPipe

  StdInPipe = {}
  StdOutPipe = {}
  StdErrPipe = {}
  
  StdInPipe = os_pipe()
  if atom(StdInPipe) then return -1 end if
  hChildStdInRd = StdInPipe[PIPE_READ_HANDLE]
  hChildStdInWr = StdInPipe[PIPE_WRITE_HANDLE]
  
  StdOutPipe = os_pipe()
  if atom(StdOutPipe) then 
    CloseAllHandles(StdInPipe)
    return -1
  end if
  hChildStdOutWr = StdOutPipe[PIPE_WRITE_HANDLE]
  hChildStdOutRd = StdOutPipe[PIPE_READ_HANDLE]
  
  StdErrPipe = os_pipe()
  if atom(StdErrPipe) then
     CloseAllHandles(StdErrPipe & StdOutPipe)
     return -1
  end if
  hChildStdErrWr = StdErrPipe[PIPE_WRITE_HANDLE]
  hChildStdErrRd = StdErrPipe[PIPE_READ_HANDLE]

  void = SetHandleInformation(StdInPipe[PIPE_WRITE_HANDLE],HANDLE_FLAG_INHERIT,0)
  void = SetHandleInformation(StdOutPipe[PIPE_READ_HANDLE],HANDLE_FLAG_INHERIT,0)
  void = SetHandleInformation(StdErrPipe[PIPE_READ_HANDLE],HANDLE_FLAG_INHERIT,0)

  return {{hChildStdInWr,hChildStdOutRd,hChildStdErrRd},
         {hChildStdInRd,hChildStdOutWr,hChildStdErrWr}}
  
end function

--WIN32 version of exec()

--**
-- Open process with command line cmd
--
-- Returns:
--   A **handle**, process handles { [[:PID]], [[:STDIN]], [[:STDOUT]], [[:STDERR]] }
--
-- Example 1:
-- <eucode>
-- object p = exec("dir", create())
-- </eucode>
--

global function exec(sequence cmd, sequence pipe)
  object fnVal
  atom hChildStdInRd,hChildStdOutWr, hChildStdErrWr,
       hChildStdInWr, hChildStdOutRd,hChildStdErrRd
  atom hChildProcess

  hChildStdInWr = pipe[1][1]
  hChildStdOutRd = pipe[1][2]
  hChildStdErrRd = pipe[1][3]
  hChildStdInRd = pipe[2][1]
  hChildStdOutWr = pipe[2][2]
  hChildStdErrWr = pipe[2][3]

  fnVal = CreateProcess(cmd,{hChildStdInRd,hChildStdOutWr,hChildStdErrWr})
  if atom(fnVal) then
    return -1
  end if
  hChildProcess = fnVal[1]
  void = close(fnVal[2])
  void = close(hChildStdInRd)
  void = close(hChildStdOutWr)
  void = close(hChildStdErrWr)
  return {hChildStdInWr, hChildStdOutRd, hChildStdErrRd, hChildProcess}

end function
